#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：设计交易模型的基础规划
# 建立日期：2023.08.29
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-08-29  第一次编写
#

from abc import ABCMeta, abstractmethod

from ..prompt import PromptManager
from ..uiacontrol import get_control_specification


class BaseModelMeta(ABCMeta):
    """交易模型元类"""

    models = {}
    model_object = {}

    def __init__(cls, name, bases, attrs):

        super(BaseModelMeta, cls).__init__(name, bases, attrs)

        if name != 'BaseModel':
            BaseModelMeta.models[attrs['name']] = cls

    def __call__(cls, client):
        # 使用唯一标识符，key或app.process（有None风险）
        if client.key not in BaseModelMeta.model_object:
            BaseModelMeta.model_object[client.key] = super(BaseModelMeta, cls).__call__(client)
        return BaseModelMeta.model_object[client.key]


class BaseModel(metaclass=BaseModelMeta):
    """交易模型基类"""

    # 交易模型的名称
    name = ''

    def __new__(cls, client):
        return object.__new__(BaseModelMeta.models[client.trademodel])

    def __init__(self, client):
        self.__element = client.mainwindow.element_info  # 加快运行速度
        self._client = client
        self._prompt = PromptManager(client)

    def _get_control(self, control_define):
        # control_define 为Client格式的字符串或字典，或者pywinauto格式的字典
        return get_control_specification(parent=self.__element, control_define=control_define)

    @abstractmethod
    def initialization(self):
        """初始化交易窗口"""
        raise NotImplementedError()

    @abstractmethod
    def reset(self):
        """复位交易窗口的功能"""
        raise NotImplementedError()


Model = BaseModel



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：执行日志记录
# 建立日期：2023.8.24
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2023-08-24  第一次编写
#

import logging
import warnings

from time import strftime, localtime
from os.path import join


class Logger(object):

    def __init__(self):
        self.__logger = logging.getLogger(__package__)
        self.__addstreamhandler()

    def __addstreamhandler(self):
        formatter = logging.Formatter(fmt='[%(asctime)s][%(name)s][%(levelname)s] %(message)s',
                                      datefmt='%Y-%m-%d %H:%M:%S')

        sh = logging.StreamHandler()
        sh.setLevel(logging.DEBUG)
        sh.setFormatter(formatter)

        self.__logger.addHandler(sh)

    def addfilehandler(self, path):
        formatter = logging.Formatter(fmt='[%(asctime)s][%(name)s][%(levelname)s] %(message)s',
                                      datefmt='%Y-%m-%d %H:%M:%S')

        file = join(path, strftime('%Y-%m-%d', localtime()) + '.log')

        fh = logging.FileHandler(file, 'a')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)

        self.__logger.addHandler(fh)

    def __set_level(self, level):
        self.__logger.setLevel(level)

    def reset(self):
        self.disable()

    def disable(self):
        self.__set_level(logging.WARNING)

    def enable(self):
        self.__set_level(logging.DEBUG)

    def debug(self, msg):
        self.__logger.debug(msg)

    def info(self, msg):
        self.__logger.info(msg)

    def warning(self, msg):
        self.__logger.warning(msg)

    def warn(self, msg):
        warnings.warn(msg)
        return self

    def error(self, msg):
        self.__logger.error(msg)

    def exception(self, msg):
        self.__logger.exception(msg)

    def critical(self, msg):
        self.__logger.critical(msg)


logger = Logger()
logger.disable()



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：各种登录引擎
# 建立日期：2023.08.16
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-08-16  第一次编写
#

from .baseengine import BaseEngine
from ..error import LoginError, TimeoutError


class VerifyCodeEngine(BaseEngine):

    name = 'VERIFYCODE'

    def __init__(self, client):
        super(VerifyCodeEngine, self).__init__(client)

    def login(self):

        for i in range(self._client.LOGIN_MAX_TIMES):
            self._prompt.close()  # 关闭可能存在的提示框或登录失败后的提示框
            self._get_control(self._client.LOGIN_USER_ID).set_text(self._client.user)  # 资金账号
            self._get_control(self._client.LOGIN_PASSWORD_ID).set_text(self._client.psw)  # 交易密码
            if self._client.second is not None:
                self._get_control(self._client.LOGIN_SECOND_ID).set_text(self._client.second)  # 第二密码
            self._get_control(self._client.LOGIN_VERIFYCODE_ID).set_text(
                self._get_control(self._client.LOGIN_VERIFYCODEIMAGE_ID).window_text()
            )  # 验证码
            self._get_control(self._client.LOGIN_SUBMIT_ID).click()  # 登录按钮
            try:
                self._client.loginwindow.wait_not('exists', timeout=self._client.LOGIN_MAX_WAIT)
                break  # 登录成功退出循环
            except TimeoutError:
                continue  # 登录不成功
        else:
            # 达到最大登录次数，登录不成功
            self._prompt.close()  # 关闭可能存在的登录错误提示框
            raise LoginError('登录不成功！')  # 登录不成功


class VerifyCodePlusEngine(BaseEngine):

    name = 'VERIFYCODEPLUS'

    def __init__(self, client):
        super(VerifyCodePlusEngine, self).__init__(client)

    def login(self):

        for i in range(self._client.LOGIN_MAX_TIMES):
            self._prompt.close()  # 关闭可能存在的提示框或登录失败后的提示框
            self._get_control(self._client.LOGIN_USER_ID).set_text(self._client.user)  # 资金账号
            self._get_control(self._client.LOGIN_PASSWORD_ID).set_text(self._client.psw)  # 交易密码
            if self._client.second is not None:
                self._get_control(self._client.LOGIN_SECOND_ID).set_text(self._client.second)  # 第二密码
            self._get_control(self._client.LOGIN_VERIFYCODE_ID).set_text(
                self._get_control(self._client.LOGIN_VERIFYCODEIMAGE_ID).refresh().window_text()
            )  # 验证码
            self._get_control(self._client.LOGIN_SUBMIT_ID).click()  # 登录按钮
            try:
                self._client.loginwindow.wait_not('exists', timeout=self._client.LOGIN_MAX_WAIT)
                break  # 登录成功退出循环
            except TimeoutError:
                continue  # 登录不成功
        else:
            # 达到最大登录次数，登录不成功
            self._prompt.close()  # 关闭可能存在的登录错误提示框
            raise LoginError('登录不成功！')  # 登录不成功


class DefaultEngine(BaseEngine):

    name = 'DEFAULT'

    def __init__(self, client):
        super(DefaultEngine, self).__init__(client)

    def login(self):

        for i in range(self._client.LOGIN_MAX_TIMES):
            self._prompt.close()  # 关闭可能存在的提示框或登录失败后的提示框
            self._get_control(self._client.LOGIN_USER_ID).set_text(self._client.user)  # 资金账号
            self._get_control(self._client.LOGIN_PASSWORD_ID).set_text(self._client.psw)  # 交易密码
            if self._client.second is not None:
                self._get_control(self._client.LOGIN_SECOND_ID).set_text(self._client.second)  # 第二密码
            self._get_control(self._client.LOGIN_SUBMIT_ID).click()  # 登录按钮
            try:
                self._client.loginwindow.wait_not('exists', timeout=self._client.LOGIN_MAX_WAIT)
                break  # 登录成功退出循环
            except TimeoutError:
                continue  # 登录不成功
        else:
            # 达到最大登录次数，登录不成功
            self._prompt.close()  # 关闭可能存在的登录错误提示框
            raise LoginError('登录不成功！')  # 登录不成功



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：设计自定义UIA控件基类
# 建立日期：2023.09.28
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2023-09-28  第一次编写
#

from ..client import get_client
from ..utils.ocr import image_to_string
from ..uiacontrol import UIAWrapper, win32structure, get_control_specification


class BaseUIAWrapper(UIAWrapper):

    _control_types = ['BaseUIA']

    def __init__(self, element_info):
        super(BaseUIAWrapper, self).__init__(element_info)
        self._client = get_client(process=element_info.process_id)
        self._prompt = self._client.prompt
        self._win32structure = win32structure

    def _get_control(self, control_define):
        # control_define 为Client格式的字符串或字典，或者pywinauto格式的字典
        return get_control_specification(self._client.window(), control_define)

    def config(self, key):  # 弹出框无法使用
        return self.element_info.config.get(key, None)

    def top_level_parent(self):
        # NOTE 官方top_level_parent()效率低且易出错，重写
        # return self._client.window().wrapper_object()  # 注意：集成环境下仍然指向客户端主窗口
        return self._client.root_window().wrapper_object()

    def standard(self):
        """返回此控件的pywinauto官方标准控件"""
        # NOTE 不要在条件中添加type和class，有可能失效
        return get_control_specification(self.element_info.parent, {'handle': self.element_info.handle})

    def own(self):  # 弹出框无法使用
        """返回此控件的另一个副本"""
        return get_control_specification(self.element_info.current_parent, self.element_info.control_define)

    def child(self, control_define):
        """返回此控件的后代规范"""
        # control_define 为Client格式的字符串或字典，或者pywinauto格式的字典
        return get_control_specification(self.element_info, control_define)

    def texts(self):
        """重写texts()"""
        rtn = [c.window_text() for c in self.descendants() if c.window_text() != '']
        return [self.window_text()] + list(map(lambda s: s.replace('\r', '').replace('\n', ''), rtn))

    def image_text(self, box=None, whitelist=None):
        """返回控件的表面可见文本"""
        if whitelist is None:
            whitelist = ''
        return image_to_string(self.capture_as_image(box), tessedit_char_whitelist=whitelist)

    def exists(self, timeout=None):
        """判断控件是否还存在"""
        return self._get_control({'handle': self.handle}).exists(timeout=timeout)


Wrapper = BaseUIAWrapper



#
# OCR（光学字符识别）tesseract 软件支持
# 官方网站：https://github.com/tesseract-ocr/tesseract
# 正确设置tesseract的安装路径，tesseract
#
#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：光学字符识别
# 建立日期：2023.04.12
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2023-04-12  第一次编写
#   2023-11-16  添加等待保存完成
#

from os import remove
from os.path import dirname, join, exists
from tempfile import NamedTemporaryFile

from .general import command

# 正确设置tesseract的安装路径，tesseract
path = join(dirname(__file__), r'Tesseract\tesseract.exe')


def file_to_text(file='', out='stdout', psm='3', **kwargs):
    if file == '':
        raise Exception('file不能为空！')

    # if 'tessedit_char_whitelist' in kwargs and kwargs['tessedit_char_whitelist'] == '':
    #    del kwargs['tessedit_char_whitelist']

    # 构造命令行
    args = f'{path} {file} {out} -l chi_sim+eng --psm {psm} -c'
    for key in kwargs:
        args = args + f' {key}=' + kwargs[key]

    # print(args)
    return command(args)


def file_to_string(file='', out='stdout', **kwargs):
    # 因为输出字符串，所以将psm设置为7，表示单行文本
    return file_to_text(file=file, out=out, psm='7', **kwargs)


def image_to_text(im=None, out='stdout', psm='3', **kwargs):
    if im is None:
        raise Exception('image不能为空！')

    w, h = im.size
    # 将delete设置为False，以免自动删除创建的临时文件
    with NamedTemporaryFile(prefix='WYH_', suffix='.png', delete=False) as f:
        # 将图像放大一倍，转换为灰度
        im.resize((w * 2, h * 2)).convert('L').save(f, format='png')

    # 等待保存完成
    while not exists(f.name):  # 等待保存完成
        pass

    # 将代码放在with外面，因为python和tesseract不能同时打开文件
    text = file_to_text(file=f.name, out=out, psm=psm, **kwargs)
    # 手动删除创建的临时文件
    if exists(f.name):
        remove(f.name)

    return text


def image_to_string(im=None, out='stdout', **kwargs):
    # 因为输出字符串，所以将psm设置为7，表示单行文本
    return image_to_text(im=im, out=out, psm='7', **kwargs)



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：客户端基类
# 建立日期：2023.07.15
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-07-15  第一次编写
#

from os.path import normpath

from pywinauto import Desktop
# from pywinauto.win32functions import BringWindowToTop
from pywinauto.application import Application, ProcessNotFoundError, WindowSpecification

from ..error import ClientConfigError


class BaseClientMeta(type):
    """客户端元类"""

    clients = []

    def __init__(cls, name, bases, attrs):

        super(BaseClientMeta, cls).__init__(name, bases, attrs)

        if name != 'BaseClient':

            cls.app = Application(backend="uia")

            if 'loginwindow' in attrs:
                # 主窗口设计成列表是为了兼容集成在行情软件中的交易客户端，例如通达信
                criterias_ = attrs['loginwindow']
                if isinstance(criterias_, dict):
                    criterias_ = [criterias_, ]
                criterias = [crit.copy() for crit in criterias_]
                # 对第0个元素完成处理，第0个元素是软件的界面，此时默认top_level_only=True，意味着只查找桌面的子项
                criterias[0]['app'] = cls.app
                criterias[0]['backend'] = 'uia'
                # 不需要对1以后的元素添加top_level_only、backend、parent等属性，因为pywinauto内部可以自动处理
                # 最后一个元素是真正的窗口，必须有control_count属性
                control_count = criterias[-1].pop('control_count', 20)
                # 构造真正的WindowSpecification对象
                cls.loginwindow = WindowSpecification(criterias[0])
                cls.loginwindow.criteria.extend(criterias[1:])
                # 添加自定义child_count属性，以识别弹窗
                cls.loginwindow.child_count = control_count
            else:
                raise ClientConfigError(f'客户端{cls}缺少关键配置<loginwindow>')

            if 'mainwindow' in attrs:
                # 主窗口设计成列表是为了兼容集成在行情软件中的交易客户端，例如通达信
                criterias_ = attrs['mainwindow']
                if isinstance(criterias_, dict):
                    criterias_ = [criterias_, ]
                criterias = [crit.copy() for crit in criterias_]
                # 对第0个元素完成处理，第0个元素是软件的界面，此时默认top_level_only=True，意味着只查找桌面的子项
                criterias[0]['app'] = cls.app
                criterias[0]['backend'] = 'uia'
                # 不需要对1以后的元素添加top_level_only、backend、parent等属性，因为pywinauto内部可以自动处理
                # 最后一个元素是真正的窗口，必须有control_count属性
                control_count = criterias[-1].pop('control_count', 4)
                # 构造真正的WindowSpecification对象
                cls.mainwindow = WindowSpecification(criterias[0])
                cls.mainwindow.criteria.extend(criterias[1:])
                # 添加自定义child_count属性，以识别弹窗
                cls.mainwindow.child_count = control_count
            else:
                raise ClientConfigError(f'客户端{cls}缺少关键配置<mainwindow>')

            # 最正确的做法是：cls.prompt = PromptManager(cls)，但会造成import冲突，所以在其元类中实现单例
            cls.prompt = None

            BaseClientMeta.clients.append(cls)

    # def __getattribute__(cls, attr):
    #     return object.__getattribute__(cls, attr)
    #
    # def __getattr__(cls, attr):
    #     if attr in BaseClient.__dict__:
    #         value = BaseClient.__dict__[attr]
    #         if isinstance(value, classmethod):
    #             # return lambda: rtn.__wrapped__(cls)  # 只适用于无参数时
    #             def wrapper(*args, **kwargs):
    #                 return value.__wrapped__(cls, *args, **kwargs)
    #             return wrapper
    #         else:
    #             return value
    #     else:
    #         raise ClientConfigError(f'客户端{cls}缺少名为<{attr}>的配置')


class BaseClient(metaclass=BaseClientMeta):
    """客户端的基类，不要直接使用"""

    #
    # 以下为固定参数，不跟随交易模板、登录引擎、交易模型的改变而不同
    #

    # 账号信息
    user = '123456'  # 资金账号、客户号、上A股东、深A股东等
    psw = '654321'  # 第一密码
    second = None  # 第二密码:通讯密码、认证口令、ukey口令等，没有时请设置为None
    account = {}  # 账户中的其他自定义信息

    # 客户端安装位置，大小写敏感，且盘符为大写，如：D:\weituo\银河证券\xiadan.exe
    path = ''

    # 客户端信息
    version = None
    name = None
    key = None  # 客户端设别符，重要关键参数，一定保持唯一

    # 客户端登录窗口和主窗口规范，重要参数，在具体的客户端设置
    # loginwindow = dict()
    # mainwindow = dict()

    # 交易模板名，默认DEFAULT模板无法满足功能要求时，应自定义模板
    tradetemplate = 'DEFAULT'
    # 登录引擎名，默认DEFAULT VERIFYCODE VERIFYCODEPLUS三种登录引擎无法满足要求时，应自定义登录引擎
    loginengine = 'DEFAULT'
    # 交易模型名，默认DEFAULT模型无法满足功能要求时，应自定义模型
    trademodel = 'DEFAULT'

    # 全局设置或控件参数
    # 设置交易的速度模式，注意：会影响同时操作的所有客户端
    TRADE_SPEED_MODE = 'fast'  # turbo（极速）、fast（快速）、defaults(默认)、slow（慢速）、dull（极慢）
    # 提示框弹出框相关
    PROMPT_TYPE_LIST = ('Pane', 'Window', 'Custom')
    PROMPT_TITLE_ID = '1365'
    PROMPT_CONTENT_ID = '1004'
    PROMPT_OKBUTTON_TITLE = '(确定|是|现在测评|我知道了|保存|立即升级).*'
    PROMPT_CANCELBUTTON_TITLE = '(取消|否|稍后测评|我知道了|以后再说).*'
    PROMPT_CLOSE_BUTTON = ('1008', '1003')  # 关闭按钮可能有不同的id

    # 结尾为 “_ID” 的参数表示控件的设别规范，完整的control_id如下：
    # control_id = 'auto_id|control_type|control_key|found_index',
    # 其中control_key为控件标识符，control_key不存在时，默认为control_type
    # auto_id，control_type，find_index应至少存在一个，应保证找到唯一控件
    # 该控件如有其他参数，采用字典形式传入

    # 控件参数和设置参数跟随交易模板、登录引擎、交易模型的不同而不同
    # 登录引擎DEFAULT、VERIFYCODE、VERIFYCODEPLUS所需要的控件规范和设置参数
    # 登录相关控件
    LOGIN_USER_ID = '1001|Edit|Editor'
    LOGIN_PASSWORD_ID = '1012|Edit|Editor'
    LOGIN_SECOND_ID = '1181|Edit|Editor'
    LOGIN_VERIFYCODE_ID = '1003|Edit|Editor'
    # 验证码图像，格式固定字典 (验证码ID，验证码显示在原图像的位置box（left, upper, right, lower），白名单whitelist，刷新按钮refresh)
    # 默认值，box=None,大小为原图像；whitelist=None，白名单为空；refresh=None刷新按钮为验证码图像本身
    LOGIN_VERIFYCODEIMAGE_ID = {
        'control': '1499|Image|Verifycode',
        'box': (None, None, None, None),
        'whitelist': '0123456789',
        'refresh': None
    }
    # 登录按钮
    LOGIN_SUBMIT_ID = '1006'
    # 登录的最大试错次数
    LOGIN_MAX_TIMES = 5
    # 每次登录的间隔时间（秒），此参数也用来判断登录是否成功，电脑性能差运行速度慢时，将此参数调高
    LOGIN_MAX_WAIT = 7

    # 交易模型DEFAULT所需要的设置参数和控件规范
    # 锁屏密码框的文本，长时间不用会被锁屏
    LOGIN_LOCKED_TEXT = '请输入您的交易密码'
    # 锁屏密码框的密码输入框
    LOGIN_LOCKPASSWORD_ID = '1039|Edit|Editor'
    # 客户端被断开连接的文本，长时间不用与服务器的连接会被断开
    LOGIN_UNLINK_TEXT = '您的请求.+已超时，可能网络忙'
    # 产品菜单
    PRODUCT_MENU_ID = {
        'control': '1001|Pane|Tabpane',
        'class_name': 'CCustomTabCtrl',
        'tabs': [
            {'name': '股票交易', 'rect': (1, 2, 64, 19), 'link': '129|Tree'},
            {'name': '开放式基金(场外)', 'rect': (65, 2, 176, 19), 'link': '240|Tree'},
            {'name': '银河理财', 'rect': (177, 2, 240, 19), 'link': '909|Tree'},
            {'name': '证券出借', 'rect': (241, 2, 304, 19), 'link': '2037|Tree'},
            {'name': '港股通', 'rect': (1, 22, 52, 39), 'link': '5199|Tree'},
            {'name': '账户管理', 'rect': (53, 22, 116, 39), 'link': '830|Tree'},
            {'name': '柜台市场产品', 'rect': (117, 22, 204, 39), 'link': '5040|Tree'}
        ]
    }
    PRODUCT_STOCK_INDEX = 0
    PRODUCT_FUND_INDEX = 1
    PRODUCT_MONEY_INDEX = 2
    PRODUCT_LEND_INDEX = 3
    PRODUCT_HONGKONG_INDEX = 4
    PRODUCT_ACCOUNT_INDEX = 5
    PRODUCT_MARKET_INDEX = 6

    # 左侧股票交易树形菜单
    LEFTMENU_STOCK_ID = '129|Tree'
    LEFTMENU_STOCK_BUY = '买入[F1]'
    LEFTMENU_STOCK_SELL = '卖出[F2]'
    LEFTMENU_STOCK_CANCEL = '撤单[F3]'
    LEFTMENU_STOCK_QUERY = '查询[F4]'

    # 左侧开放式基金树形菜单
    # 默认无此功能，可自行添加

    # 股票交易界面，买入卖出使用同一下单界面
    STOCK_CODE_ID = '1032|Edit|Editor'
    STOCK_PRICE_ID = '1033|Edit|Editor'
    STOCK_COUNT_ID = '1034|Edit|Editor'
    STOCK_TRADE_ID = '1006'
    # 委托买卖确认框标题，支持正则表达式
    ENTRUST_CONFIRMBOX_TITLE = '委托确认'
    # 委托买卖提示框标题，支持正则表达式
    ENTRUST_PROMPTBOX_TITLE = r'提示.*'
    # 委托买卖成功后的提示内容信息正则
    ENTRUST_SUCCESS_CONTENT = r'您的(买入|卖出)委托已成功提交，合同编号：(\d+)。'
    # 委托买卖成功后，“合同编号”位于上述表达式中的组号
    ENTRUST_SUCCESS_SERIALINDEX = 2

    # 表格相关
    GRID_DEFAULT_ID = {
        'control': '1047|Pane|GridCSV|1',
        'headHeight': 24,
        'lineHeight': 23,
        'offset': 6,
        'saveto': '|Window|Prompt',
        'savetofile': '1152|Edit'
    }

    # 撤单界面
    CANCLE_SUBMIT_ID = '1099'
    # 委托撤单确认框内容，支持正则表达式
    ENTRUST_CANCELPROMPT_TITLE = r'提示信息'

    #
    # 以下代码不要修改，不要修改！
    #
    # 启动真实客户端
    @classmethod
    def connect(cls):
        try:
            cls.app.connect(path=normpath(cls.path), timeout=0.6)
        except ProcessNotFoundError:
            cls.app.start(normpath(cls.path))  # app.start已内置等待机制
            # 但同花顺的下单程序集成在行情软件中，start()等待机制会失效,pywinauto会弹出警告,使用下列代码完成等待
            # NOTE 不要勾选自动登录，如需要，使用try...except...兼容自动登录
            WindowSpecification(cls.loginwindow.criteria[0]).wait("ready", timeout=15)  # 超时软件无法打开

        try:
            # 等待安静下来，对客户端的操作应该等待其上一个动作完全完成
            cls.app.wait_cpu_usage_lower(threshold=1.5, timeout=10)
        except RuntimeError:
            pass

        return cls

    @classmethod
    def window(cls):
        # 始终返回正确的主窗口，不要使用Application.top_window(),Application.active()来设置主窗口
        # 设置timeout=0，因应用场景非此即彼，并非捕捉模式，以此加快运行速度
        # 不用改变存在的判断顺序
        # return cls.mainwindow if cls.mainwindow.exists(timeout=0) else cls.loginwindow
        return cls.loginwindow if cls.loginwindow.exists(timeout=0) else cls.mainwindow

    @classmethod
    def root_window(cls):
        # 始终返回正确的根窗口，集成环境下的根窗口不是主窗口，即使集成环境下两个窗口均未打开，也能正确返回根窗口
        return WindowSpecification(cls.window().criteria[0])

    @classmethod
    def active(cls):
        # 激活客户端，NOTE：is_active()会误判
        root_window = cls.root_window()

        if root_window.is_dialog() and root_window.is_minimized():
            cls.taskbaricon().click()
        else:
            # BringWindowToTop(root_window.handle)
            root_window.set_focus()

        return cls

    @classmethod
    def taskbaricon(cls):
        # 客户端的任务栏图标按钮，未使用taskbar模块，是因为该模块仍处于实验阶段（pywinauto-0.6.8）
        # FIXME 未来使用taskbar模块
        return Desktop(backend='uia').window(class_name='Shell_TrayWnd', found_index=0).child_window(
            auto_id=normpath(cls.path),
            control_type='Button'
        )

    @classmethod
    def hook(cls):
        # 一个钩子，在激活（active）客户端后被调用，在子类重写它完成想要的功能
        # 或者实现在集成环境下打开内嵌的客户端，如下代码：
        # if not cls.window().exists(timeout=0):
        #     root_window = WindowSpecification(cls.loginwindow.criteria[0])
        #     # 此处用操作root_window打开内嵌客户端
        #     ...
        #     cls.loginwindow.wait('ready', timeout=15)

        return cls

    @classmethod
    def close(cls):
        # 关闭客户端
        try:
            cls.app.connect(path=normpath(cls.path), timeout=0.6).kill()
        except ProcessNotFoundError:
            pass


Client = BaseClient



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：工具库
# 建立日期：2022.11.17
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-11-17  第一次编写
#

from subprocess import run, PIPE, STDOUT
from time import sleep


def command(cmd: str) -> str:
    # 发送并执行cmd指令
    return run(cmd, stdout=PIPE, stderr=STDOUT, shell=True, timeout=None,
               encoding='gbk').stdout.replace('\r', '').replace('\n\n', '\n').strip()


# def logger(*args, sep=' ', end='\n'):
#    # 打印日志
#    path = os.getcwd() + r'\log'
#    if not os.path.exists(path): os.mkdir(path)
#    file = path + '\\' + strftime("%Y-%m-%d", localtime()) + '.log'
#    log = '[' + strftime("%H:%M:%S", localtime()) + '] ' + sep.join(args) + end
#    with open(file, 'a') as f: f.write(log)
#    print('[' + strftime("%H:%M:%S", localtime()) + ']', *args, sep=sep, end=end)


def wait(s):
    sleep(s)



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：各种登录引擎
# 建立日期：2023.08.16
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-08-16  第一次编写
#

# from pywinauto.timings import TimeoutError

# from .baseengine import BaseEngine
# from ..error import LoginError

# 将来更多登录引擎



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：各种客户端定义
# 建立日期：2023.07.15
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-07-15  第一次编写
#

from os.path import dirname, join
from .baseclient import BaseClient


class YH(BaseClient):
    """银河证券客户端"""
    # 账号信息
    user = '123456'   # 资金账号，目前只支持资金账号登录
    psw = '123456'   # 密码

    # 客户端安装位置
    path = join(dirname(__file__), r'银河证券\xiadan.exe')

    # 客户端信息
    version = '5.0'
    name = '银河双子星'
    key = 'yh'

    # 客户端登录窗口和主窗口规范,只要能区别即可，没必要都填写（除control_count外），使用inspect.exe查看，参数如下：
    # title：            有这个标题的窗口，inspect.exe下的name属性
    # title_re：         标题与此正则表达式匹配的窗口
    # class_name：       具有此窗口类的窗口
    # class_name_re：    类与此正则表达式匹配的窗口
    # control_type：     具有此控件类型的窗口
    # auto_id：          具有此自动化ID的窗口
    # control_count:     该窗口在无任何弹窗时的子项数，*必填项*
    loginwindow = dict(title='用户登录', control_count=20)
    mainwindow = dict(title='网上股票交易系统5.0', control_count=4)

    # 登录引擎名
    loginengine = 'VERIFYCODEPLUS'



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：各种交易模型
# 建立日期：2023.08.29
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-08-29  第一次编写
#

import re

from .basemodel import BaseModel
from ..logger import logger
from ..error import StockCodeError, StockPriceError, StockCountError, TradeFailFError


class DefaultModel(BaseModel):
    """默认交易模型"""

    name = 'DEFAULT'

    def __init__(self, client):
        super(DefaultModel, self).__init__(client)

    def __product_menu_select(self, item):
        """选择产品"""
        self._get_control(self._client.PRODUCT_MENU_ID).select(item)

    def __left_menu_select(self, menu, item):
        """左侧树形菜单功能选择"""
        self._get_control(menu).get_item(item).select()

    def __product_item_select(self, product, item):
        """同时选择产品和功能"""
        self._get_control(self._client.PRODUCT_MENU_ID).select(product).get_item(item).select()

    def __verify_stock_code(self, code):
        """检查证券代码"""
        if re.match(r'^\d{6}$', code) is None:
            raise StockCodeError('证券代码必须为6位数字')

        self._get_control(self._client.STOCK_CODE_ID).set_text(code)

    def __verify_stock_price(self, price):
        """
        检查价格是否正确，此方法并没有对价格的最小单位进行限制
        """
        # 价格为空时，默认当前买入卖出价
        if price is None or price == '':
            return

        # 没有对小数点后的位数进行限制
        if re.match(r'^(([1-9]\d*)|0)(\.\d*)?$', str(price)) is None:
            raise StockPriceError('价格必须为规范的数字型')

        self._get_control(self._client.STOCK_PRICE_ID).set_text(str(price))

    def __verify_stock_count(self, count):
        """
        检查买卖数量是否正确，此方法并没有对数量是否满足手（或张）的倍数进行验证
        """
        if count is None or re.match(r'^[1-9]\d*$', str(count)) is None:
            raise StockCountError('买卖数量必须为规范的数字型或不为0')

        self._get_control(self._client.STOCK_COUNT_ID).set_text(str(count))

    def __verify_stock_trade(self):
        """验证交易是否正确"""
        pane = self._prompt.tooltip(title=self._client.ENTRUST_CONFIRMBOX_TITLE)
        if pane is not None:
            pane.ok()

        pane = self._prompt.tooltip(title=self._client.ENTRUST_PROMPTBOX_TITLE)
        if pane is not None:
            content = pane.content()
            pane.close()
            mat = re.match(self._client.ENTRUST_SUCCESS_CONTENT, content)
            if mat is not None:
                return mat.group(self._client.ENTRUST_SUCCESS_SERIALINDEX)  # 成功则返回合同编号
            else:
                raise TradeFailFError(content)

        logger.warn('捕捉不到委托合同编号，可能客户端关闭了委托成功提示框')
        return ''

    def __trade(self, code='', price=None, count=None):
        # 买入和卖出共享一个下单界面
        self.__verify_stock_code(code)  # 证券代码
        self.__verify_stock_price(price)  # 买卖价格
        self.__verify_stock_count(count)  # 买卖数量
        # 买入/卖出
        self._prompt.close()  # 关闭可能存在的弹出框
        self._get_control(self._client.STOCK_TRADE_ID).click()
        return self.__verify_stock_trade()

    def initialization(self):
        """初始化交易窗口，暂不支持精简模式"""
        if self._prompt.exists(text=self._client.LOGIN_LOCKED_TEXT):
            # 客户端被锁屏
            pane = self._prompt.tooltip(text=self._client.LOGIN_LOCKED_TEXT)
            pane.child(self._client.LOGIN_LOCKPASSWORD_ID).set_text(self._client.psw)
            pane.ok()

        if self._prompt.exists(text=self._client.LOGIN_UNLINK_TEXT):
            # 客户端长时间闲置，连接被断开
            self._prompt.tooltip(text=self._client.LOGIN_UNLINK_TEXT).ok()

        # 关闭其他弹窗
        self._prompt.close()

    def reset(self):
        """复位交易窗口的功能"""
        # 复位至股票买入
        # self.__product_item_select(self._client.PRODUCT_STOCK_INDEX, [self._client.LEFTMENU_STOCK_BUY])
        pass

    def buy(self, code='', price=None, count=None, **kwargs):
        """执行买入操作"""
        self.__product_item_select(self._client.PRODUCT_STOCK_INDEX, [self._client.LEFTMENU_STOCK_BUY])
        return self.__trade(code=code, price=price, count=count)

    def sell(self, code='', price=None, count=None, **kwargs):
        """执行卖出操作"""
        self.__product_item_select(self._client.PRODUCT_STOCK_INDEX, [self._client.LEFTMENU_STOCK_SELL])
        return self.__trade(code=code, price=price, count=count)  # 买入卖出使用一个下单界面

    def cancel(self, **kwargs):
        """执行撤单操作"""
        self.__product_item_select(self._client.PRODUCT_STOCK_INDEX, [self._client.LEFTMENU_STOCK_CANCEL])

        rtn = []
        for item in self._get_control(self._client.GRID_DEFAULT_ID).refresh().items(**kwargs):
            item.select()
            rtn.append(item)

        if rtn:
            # 关闭可能存在的弹出框
            self._prompt.close()
            # 点击撤单按钮, 不需要等待撤单按钮可用，底层已有等待机制 .wait('enabled')
            self._get_control(self._client.CANCLE_SUBMIT_ID).click()
            # 弹出确认对话框
            pane = self._prompt.tooltip(text=self._client.ENTRUST_CANCELPROMPT_TITLE)  # 设置正确的参数
            if pane is not None:
                pane.ok()
            # 关闭可能的提示对话框
            self._prompt.start_monitor(delay=1.5)

        return rtn

    def query(self, target, **kwargs):
        self.__product_item_select(self._client.PRODUCT_STOCK_INDEX, [self._client.LEFTMENU_STOCK_QUERY, target])
        return self._get_control(self._client.GRID_DEFAULT_ID).refresh()



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：券商客户端库
# 建立日期：2023.08.15
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 项目软件特色：
# 该项目是始于2023年初的一个实验项目，由于个人的原因而编写，但笔者上次编写代码还是近20年前，主要使用C和汇编编写一些很底层
# 的项目。所以，该项目是笔者第一个python项目，错误在所难免。该项目通过使用python第三方库pywinauto，实现对券商客户端的
# 自动化操作测试，包括自动登录、验证码识别、买卖、撤单、查询等功能。该项目深度应用pywinauto库，代码中有许多pywinauto库
# 的应用技巧，包括该库存在的BUG也已在代码中标明。该项目具有以下特点：
# * 该项目采用一种“客户端驱动型工厂模式”设计，或者叫“客户定制式工厂模式”。所以，从理论上讲，该项目具备了支持所有券商客户端
#   及其未来版本的能力。
# * 由于交易的严谨性和严肃性，所有的操作均有返回值，要么成功要么失败，不会因运行时错误而“卡”在半路。
# * 自动化软件测试受内外环境的影响较大，该项目以最大的可能减少内外环境的变化对软件自动化的影响。
# * 由于采用“客户端驱动型工厂模式”，所以项目可扩展性高、可根据不同的券商版本制作不同的交易模型也可以制作不同的部件适应不同的
#   场景，如制作不同的登录引擎以适应不同的登录方式。
#

from . import brokersa, brokersb  # 导入模块以注册模块中的所有券商客户端
from ..error import ClientConfigError


def get_client(process):

    clients = brokersa.YH.clients

    for client in reversed(clients):
        if client.app.process is not None and client.app.process == process:
            return client
    else:
        raise ClientConfigError(f'找不到客户端(process={process})')



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：邮件发送
# 建立日期：2023.09.28
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2023-10-13  第一次编写
#   2023-11-29  添加登录函数
#
"""
 使用方法：

 from pytradecn import email

 email.login(server='smtp.qq.com', user='41715399@qq.com', password='您的密码')
 mail = email.Mail(subject='这是一封邮件')
 mail.content(text='邮件内容').send(receivers='41715399@qq.com', sender='我是发送者')

"""

from os.path import basename

from smtplib import SMTP_SSL, SMTPException
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.header import Header
from email.utils import formataddr


def login(server, user, password):

    Mail.SERVER = server
    Mail.USER = user
    Mail.PASSWORD = password


class Mail:

    SERVER = 'smtp.qq.com'
    USER = '41715399@qq.com'
    PASSWORD = '您的密码'

    def __init__(self, subject='未设置主题'):
        self.__subject = subject
        self.__mail = MIMEMultipart('related')

    def content(self, text='这是一封邮件', images=None, atts=None):
        """邮件正文
        text:  邮件正文内容，默认为空字符，其值可以为普通文本或HTML
               text = '这是我的正文'  # 普通的文本
               text = '<p><b>这是我的正文</b></p>'  # HTML文本
        images:图像列表，默认为None，其值可以单个硬盘图像、单个文件对象、或他们组成的元组
               images = 'd:/123.jpg'  # 字符串形式的单文件图像
               images = file对象  # 单图像文件对象 open()已打开的文件对象或PIL文件对象
               images = ('d:/123.jpg', file, 'd:/456.jpg')  # 他们组成的元组
        atts:  附件或附件列表，默认为None，其值可以为单个硬盘文件或他们组成的元组
               atts = 'd:/123.docx'  # 字符串形式的单个文件
               atts = ('d:/123.docx', 'd:/456.docx')  # 附件字符串组成的元组
        """
        if images is not None and not isinstance(images, tuple):
            images = (images,)

        if atts is not None and isinstance(atts, str):
            atts = (atts,)

        # 邮件正文
        if images is not None:
            for i, image in enumerate(images, start=1):
                text = text + rf'<p><img src="cid:image{i}"></p>'
                if isinstance(image, str):
                    with open(image, 'rb') as f:
                        # img = MIMEImage(f.read(), _subtype=False)
                        img = MIMEImage(f.read())
                else:
                    img = image
                img.add_header('Content-ID', f'<image{i}>')
                self.__mail.attach(img)

        if atts is not None:
            for file in atts:
                att = MIMEText(open(file, 'rb').read(), 'base64', 'utf-8')
                att["Content-Type"] = 'application/octet-stream'
                # att["Content-Disposition"] = 'attachment; filename="' + basename(file) + '"'
                att.add_header('Content-Disposition', 'attachment', filename=('gbk', '', basename(file)))
                self.__mail.attach(att)

        # 必须将text放在最后
        if text != '':
            self.__mail.attach(MIMEText(text, 'html', 'utf-8'))

        return self

    def send(self, receivers, sender=None):
        """发送邮件
        receivers: 邮件接收者，必填项，其值可以为单个字符串地址、字符串地址列表
               receivers = '12345@qq.com'  # 字符串表示的邮件地址
               receivers = ['12345@qq.com', '67890@qq.com']  # 字符串邮件地址列表
        sender: 邮件发送者，默认为 41715399@qq.com，其值可以为名称字符串、或名称和地址组成的二元组
               sender = '张三'  # 字符串，发送者的机构或名称
               sender = ('张三', '12345@qq.com')  # 发送者名称（机构）和地址组成的二元组
        """
        if isinstance(receivers, str):
            receivers = [receivers, ]

        if sender is None:
            sender = Mail.USER

        if isinstance(sender, str):
            sender = (sender, Mail.USER)

        # 邮件头
        # self.__mail['From'] = formataddr((Header(sender[0], 'utf-8').encode(), sender[1]))
        self.__mail['From'] = formataddr(sender)
        self.__mail['To'] = ';'.join(receivers)
        self.__mail['Subject'] = Header(self.__subject, 'utf-8')

        # 发送邮件
        try:
            smtp = SMTP_SSL(Mail.SERVER)
            # 我们用set_debuglevel(1)就可以打印出和SMTP服务器交互的所有信息。
            # smtp.set_debuglevel(1)
            smtp.login(Mail.USER, Mail.PASSWORD)
            smtp.sendmail(sender[1], receivers, self.__mail.as_string())
            smtp.quit()
        except SMTPException:
            pass



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：UIA控件规范
# 建立日期：2023.08.05
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2023-08-05  第一次编写
#

import re

from pywinauto import win32structures
from pywinauto.application import WindowSpecification
from pywinauto.controls.uiawrapper import UIAWrapper
from pywinauto.findbestmatch import MatchError
from pywinauto.controls import InvalidWindowHandle, InvalidElement
from pywinauto.timings import Timings, wait_until, wait_until_passes, TimeoutError

from .error import ClientConfigError, ElementAmbiguousError, ElementNotFoundError

win32structure = win32structures


def find_elements(class_name=None,
                  class_name_re=None,
                  parent=None,
                  title=None,
                  title_re=None,
                  visible_only=True,
                  enabled_only=True,  # 默认为Ture
                  # best_match=None,  # 此场景没什么用
                  handle=None,
                  ctrl_index=None,
                  found_index=None,
                  predicate_func=None,
                  # active_only=False,  # 此场景无用
                  control_id=None,
                  control_type=None,
                  auto_id=None,
                  framework_id=None,
                  depth=None,
                  # 自定义参数
                  control_key=None,
                  control_define=None,
                  top_parent=None,
                  **config
                  ):
    """查找元素"""
    elements = parent.descendants(class_name=class_name,
                                  control_type=control_type,
                                  cache_enable=True,
                                  depth=depth
                                  )

    # 按顺序过滤，最大可能减少循环次数
    if ctrl_index is not None:
        elements = elements[ctrl_index:ctrl_index+1] if len(elements) > ctrl_index else []

    if handle is not None and elements:  # 不用采用官方的方法
        elements = [elem for elem in elements if elem.handle == handle]

    if auto_id is not None and elements:
        elements = [elem for elem in elements if elem.automation_id == auto_id]

    if control_id is not None and elements:
        elements = [elem for elem in elements if elem.control_id == control_id]

    if framework_id is not None and elements:
        elements = [elem for elem in elements if elem.framework_id == framework_id]

    if title is not None and elements:
        # TODO: some magic is happenning here
        elements = [elem for elem in elements if elem.rich_text == title]
    elif title_re is not None and elements:
        title_regex = re.compile(title_re)

        def _title_match(w):
            """Match a window title to the regexp"""
            t = w.rich_text
            if t is not None:
                return title_regex.match(t)
            return False

        elements = [elem for elem in elements if _title_match(elem)]

    # if class_name is not None and elements:
    #     elements = [elem for elem in elements if elem.class_name == class_name]

    if class_name_re is not None and elements:
        class_name_regex = re.compile(class_name_re)
        elements = [elem for elem in elements if class_name_regex.match(elem.class_name)]

    if visible_only and elements:
        elements = [elem for elem in elements if elem.visible]

    if enabled_only and elements:
        elements = [elem for elem in elements if elem.enabled]

    # 倒数第二个过滤条件
    if predicate_func is not None and elements:
        elements = [elem for elem in elements if predicate_func(elem)]

    # 最后一个过滤条件
    if found_index is not None:
        elements = elements[found_index:found_index+1] if len(elements) > found_index else []

    # 设置自定义参数
    for ele in elements:
        ele.control_key = control_key or ele.control_type  # control_key if control_key isnot None else ele.control_type
        ele.control_define = control_define
        ele.top_parent = top_parent
        ele.current_parent = parent
        ele.config = config

    return elements


class UIAControlSpecification(object):

    def __init__(self, criteria):

        self.__criteria = criteria

        self.__wait_criteria_map = {
            'exists': ('exists',),
            'visible': ('is_visible',),
            'enabled': ('is_enabled',),
            'ready': ('is_visible', 'is_enabled',),
            'active': ('is_active',),
        }

    def __getattribute__(self, attr):
        return object.__getattribute__(self, attr)

    def __getattr__(self, attr):
        return getattr(self.wrapper_object(), attr)

    def __get_ctrl(self):

        elements = find_elements(**self.__criteria)

        if not elements:
            raise ElementNotFoundError('找不到控件:{0}'.format(self.__criteria['control_define']))

        if len(elements) > 1:
            exception = ElementAmbiguousError('找到{0}个控件，在此条件下:{1}'.format(
                len(elements),
                str(self.__criteria['control_define']),
            ))
            exception.elements = elements
            raise exception

        return UIAWrapper(elements[0])

    def __resolve_control(self, timeout=None, retry_interval=None):

        # 不要设置为默认值
        if timeout is None:
            timeout = Timings.window_find_timeout
        if retry_interval is None:
            retry_interval = Timings.window_find_retry

        try:
            return wait_until_passes(
                timeout,
                retry_interval,
                self.__get_ctrl,
                (ElementNotFoundError, ElementAmbiguousError, MatchError, InvalidWindowHandle, InvalidElement)
                )
        except TimeoutError as err:
            raise ClientConfigError(str(err.original_exception))

    def wrapper_object(self):
        return self.__resolve_control()

    def exists(self, timeout=None, retry_interval=None):
        """判断控件是否存在，默认非捕捉模式"""

        # 不要设置为默认值，时间会更改
        if timeout is None:
            # timeout = Timings.exists_timeout
            timeout = 0
        if retry_interval is None:
            retry_interval = Timings.exists_retry

        # 默认值visible_only=True, enabled_only=True, 官方关闭这两个参数主要针对窗口且有其普遍道理
        # 不用关闭这两个参数，盈余规范参数保持一致
        # visible_only = self.__criteria.get('visible_only', True)
        # enabled_only = self.__criteria.get('enabled_only', True)
        #
        # self.__criteria['visible_only'] = False
        # self.__criteria['enabled_only'] = False

        try:
            self.__resolve_control(timeout, retry_interval)
            # self.__criteria['visible_only'] = visible_only
            # self.__criteria['enabled_only'] = enabled_only
            return True
        except ClientConfigError:  # 这里跟__resolve_control抛出的错误一致
            # self.__criteria['visible_only'] = visible_only
            # self.__criteria['enabled_only'] = enabled_only
            return False

    def __parse_wait_args(self, wait_conditions, timeout, retry_interval):

        # 解析等待条件，直接使用官方代码
        if timeout is None:
            timeout = Timings.window_find_timeout
        if retry_interval is None:
            retry_interval = Timings.window_find_retry

        # 转换成小写
        wait_for = wait_conditions.lower()

        # 集合具有去重功能
        unique_check_names = set()
        # 将字符串解析成列表
        wait_criteria_names = wait_for.split()
        # 遍历等待列表
        for criteria_name in wait_criteria_names:
            try:
                # 从映射表中获得方法元组
                check_methods = self.__wait_criteria_map[criteria_name]
            except KeyError:
                # Invalid check name in the wait_for
                raise SyntaxError('意外的条件 - %s' % criteria_name)
            else:
                # 将方法元组添加到集合中，注：集合会解包元组，并去掉重复项
                unique_check_names.update(check_methods)

        # 最终的unique_check_names如下：
        # unique_check_names = set(['is_enabled', 'is_active', 'is_visible', 'Exists'])
        return unique_check_names, timeout, retry_interval

    def __check_all_conditions(self, check_names, retry_interval):

        # 检查所有的条件，直接使用官方代码

        # 遍历所有的检查方法
        for check_name in check_names:
            # timeout = retry_interval because the timeout is handled at higher level
            if check_name == 'exists':  # 如果是exists()
                check = getattr(self, check_name)
                if not check(retry_interval, float(retry_interval) // 2):
                    return False  # 不成功，直接返回False
                else:
                    continue  # 成功则检查下一个条件
            try:
                # 获得控件包装器
                ctrl = self.__resolve_control(retry_interval, float(retry_interval) // 2)
                check = getattr(ctrl, check_name)  # 获得方法
            except ClientConfigError:  # 这里跟__resolve_control抛出的错误一致
                # 控件不存在，直接返回False. 所以'enabled', 'active', 'visible', 'ready'会同时检查控件是否存在
                return False
            else:
                if not check():
                    # 条件不满足，直接返回False
                    return False
            # 满足条件，继续循环下一个判断

        # 所有的条件均满足，返回True
        return True

    def wait(self, wait_for, timeout=None, retry_interval=None):
        # 直接使用官方代码
        check_method_names, timeout, retry_interval = self.__parse_wait_args(wait_for, timeout, retry_interval)
        wait_until(timeout, retry_interval,
                   lambda: self.__check_all_conditions(check_method_names, retry_interval))

        # 返回控件包装器
        return self.wrapper_object()

    def wait_not(self, wait_for_not, timeout=None, retry_interval=None):
        # 直接使用官方代码
        check_method_names, timeout, retry_interval = self.__parse_wait_args(wait_for_not, timeout, retry_interval)
        wait_until(timeout, retry_interval,
                   lambda: not self.__check_all_conditions(check_method_names, retry_interval))


def str2dict(control_str: str) -> dict:

    criteria = {}

    args = control_str.split('|')
    # 1.设置found_index参数
    criteria['found_index'] = None if len(args) < 4 or args[3] == '' else int(args[3])
    # 2.设置predicate_func参数，以实现控件设别
    criteria['control_key'] = None if len(args) < 3 or args[2] == '' else args[2]
    # 3.设置control_type参数
    criteria['control_type'] = None if len(args) < 2 or args[1] == '' else args[1]
    # 4.设置auto_id参数，字符串始终有一个
    criteria['auto_id'] = None if args[0] == '' else args[0]

    return criteria


def criteria_parser(control_define: (str, dict)) -> dict:

    if isinstance(control_define, str):
        criteria = str2dict(control_define)
    elif isinstance(control_define, dict) and 'control' in control_define:
        criteria = str2dict(control_define['control'])
        criteria.update(control_define)
        criteria.pop('control', None)
    else:
        criteria = control_define.copy()

    criteria['control_define'] = control_define

    return criteria


def get_control_specification(parent, control_define, top_parent=None):
    # 必须包含parent参数
    if isinstance(parent, (WindowSpecification, UIAControlSpecification, UIAWrapper)):
        parent = parent.element_info

    if top_parent is not None:
        if isinstance(top_parent, (WindowSpecification, UIAControlSpecification, UIAWrapper)):
            top_parent = top_parent.element_info
    else:
        top_parent = parent

    criteria = criteria_parser(control_define)
    criteria['parent'] = parent
    criteria['top_parent'] = top_parent

    return UIAControlSpecification(criteria)



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：各种自定义控件
# 建立日期：2023.08.10
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-08-10  第一次编写
#

from .baseuiawrapper import BaseUIAWrapper


class EditorWrapper(BaseUIAWrapper):
    """编辑器控件，实现统一接口"""

    _control_types = ['Editor']

    def __init__(self, elem):
        super(EditorWrapper, self).__init__(elem)

    def set_edit_text(self, text):
        # 直接使用self.xxx_input()等模拟鼠标键盘事件，会在UIAElementInfo()出现"COMError -2147220991, '事件无法调用任何订户'错误"
        # 原因可能与操作系统相关的权限控制有关，但pywinauto(0.6.8)不会报错，同时将所有的element属性置为None。
        # 但WindowSpecification.xxx_input()不会有问题，WindowSpecification每次操作会新建不同的包装器对象，
        # 故采用以下方法，每次使用不同的包装器跳过此问题
        self.standard().double_click_input()
        # self.standard().type_keys('{DELETE}')
        self.standard().type_keys(str(text))

        return self

    set_text = set_edit_text


class VerifycodeWrapper(BaseUIAWrapper):

    _control_types = ['Verifycode']

    def __init__(self, elem):
        """Initialize the control"""
        super(VerifycodeWrapper, self).__init__(elem)

        box = self.config('box')
        if box is not None:
            self.__box = self._win32structure.RECT()
            self.__box.left = self.rectangle().left if box[0] is None else self.rectangle().left + int(box[0])
            self.__box.top = self.rectangle().top if box[1] is None else self.rectangle().top + int(box[1])
            self.__box.right = self.rectangle().right if box[2] is None else self.rectangle().left + int(box[2])
            self.__box.bottom = self.rectangle().bottom if box[3] is None else self.rectangle().top + int(box[3])
        else:
            self.__box = self.rectangle()

        self.__whitelist = self.config('whitelist')

    def window_text(self):
        """返回Image控件的窗口文本"""
        # NOTE 这里的box相对其父元素,没有采用crop()是考虑其保存更大图像
        return self.image_text(box=self.__box, whitelist=self.__whitelist)

    def refresh(self):
        if self.config('refresh') is None:
            self.click_input(coords=(self.__box.mid_point().x - self.rectangle().left,
                                     self.__box.mid_point().y - self.rectangle().top
                                     )
                             )
        else:
            self._get_control(self.config('refresh')).click_input()

        # 验证码刷新后，原控件已失效，返回其副本
        return self.own()


class ImagedigitWrapper(BaseUIAWrapper):

    _control_types = ['Imagedigit']

    def __init__(self, elem):
        """Initialize the control"""
        super(ImagedigitWrapper, self).__init__(elem)

    def window_text(self):
        return self.image_text(whitelist='0123456789.')


class TabPaneWrapper(BaseUIAWrapper):
    """Pane类型的Tab控件"""

    _control_types = ['Tabpane']

    def __init__(self, elem):
        super(TabPaneWrapper, self).__init__(elem)

        self.__tabs = [] if self.config('tabs') is None else self.config('tabs')

    def select(self, *item):
        # 参数*item写这样，是为了让PyCharm停止警告与基类签名不匹配，不是低级错误
        # 目前只支持index
        point = self._win32structure.RECT(*self.__tabs[item[0]].get('rect')).mid_point()
        self.click_input(coords=(point.x, point.y))

        link = self.__tabs[item[0]].get('link', None)
        if link is not None:
            return self._get_control(link)
        else:
            return self

    def tab_count(self):
        return len(self.__tabs)

    def texts(self):
        return [tab['name'] for tab in self.__tabs]



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：设计总体架构的模板规划
# 建立日期：2023.08.20
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2023-08-20  第一次编写
#   2024-01-25  0.0.4版本都改成由__getattribute__实现，去掉编写模板的步骤
#
"""
    模板就象是汽车的总装车间，模板基类用来完成交易模板的基础行为，模板只用来定义功能而不实现功能，功能的实现应有交易模型（model）完成。
    0.0.4版本后只保留基础模板，功能的定义和实现均在模型中完成
"""

# from abc import ABCMeta, abstractmethod
# from functools import wraps

from pywinauto.timings import Timings
from pywinauto.application import AppStartError

from ..client.baseclient import BaseClientMeta
from ..prompt import PromptManager
from ..engine.baseengine import BaseEngine
from ..model.basemodel import BaseModel
from ..logger import logger
from ..error import ClientConfigError, TimeoutError


class BaseTemplateMeta(type):
    """交易模板元类"""

    # templates = {}

    def __init__(cls, name, bases, attrs):

        super(BaseTemplateMeta, cls).__init__(name, bases, attrs)

        # if name != 'BaseTemplate':
        #     BaseTemplateMeta.templates[attrs['name']] = cls

    def __call__(cls, client=None, user=None, psw=None, second=None, **account):
        client = client or BaseClientMeta.clients[-1]  # if client is None else client
        client.user = user or client.user  # if user is not None else client.user
        client.psw = psw or client.psw  # if psw is not None else client.psw
        client.second = second or client.second  # if second is not None else client.second
        client.account.update(account)
        return super(BaseTemplateMeta, cls).__call__(client)


class BaseTemplate(metaclass=BaseTemplateMeta):
    """
    交易模板的基类，有4个功能在其子类中必须有定义，分别是buy（买入）、sell（卖出）、cancel（撤单）、query（查询），任何在子类中定义
    的功能都必须添加@BaseTemplate.connect修饰器才能正常工作。在子类中self._client用于访问客户端，self._prompt用于访问弹窗管理
    器，模板基类是唯一对外接口，外部访问时使用Trader()访问，下面是在您的项目中的访问方法：

    """

    # name = ''  # 交易模板的名称

    # def __new__(cls, client):
    #     return object.__new__(BaseTemplateMeta.templates[client.tradetemplate])

    def __init__(self, client):
        self._client = client
        self._prompt = PromptManager(client)
        getattr(Timings, client.TRADE_SPEED_MODE)()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # if exc_type is not None:
        #    logger.error(''.join(traceback.format_exception(exc_type, exc_val, exc_tb)))
        self.close()

    def __getattribute__(self, item):
        return object.__getattribute__(self, item)

    def __getattr__(self, item):
        def wrapper(*args, **kwargs):
            try:
                self.__connect()
                return True, getattr(self._model, item)(*args, **kwargs)
            except Exception as err:
                logger.exception(str(err))
                return False, str(err)

        return wrapper

    def close(self):
        self._prompt.stop_monitor()
        self._client.close()

    def __login(self):
        if self._client.window() is self._client.loginwindow:
            # 用户未登录
            BaseEngine(self._client).login()
            self._client.mainwindow.wait('ready', timeout=15)  # 等待交易主窗口准备好
            self._prompt.start_monitor(delay=5)  # 关闭自动弹出的提示框
            BaseModel.model_object.pop(self._client.key, None)  # 建立新对象
            self._model = BaseModel(self._client)
        else:
            # 用户已登录
            self._model = BaseModel(self._client)
            self._model.initialization()  # 初始化交易窗口

        self._model.reset()

    def __hook(self):
        self._client.hook()

    def __active(self):
        self._client.active()

    def __setapp(self):
        try:
            self._client.connect()
        except (AppStartError, TimeoutError):
            raise ClientConfigError(f'无法启动客户端，可能路径拼写错误：{self._client.path}')

    def __unlock(self):
        """软件的自动化依赖电脑在登录的情况下"""
        # if win32gui.GetForegroundWindow() == 0:
        #    raise ScreenLockedError('屏幕被锁定')  # 操作系统限制，无法用软件解锁电脑
        # return self
        pass

    def __connect(self):
        # 1.电脑屏幕是否被锁定
        self.__unlock()
        # 2.启动应用程序
        self.__setapp()
        # 3.激活应用程序
        self.__active()
        # 4.调用钩子
        self.__hook()
        # 5.登录应用程序
        self.__login()

    @staticmethod
    def connect(func):  # 保留以兼容0.0.4版本以前的插件
        pass

    # @staticmethod
    # def connect(func):
    #     @wraps(func)
    #     def wrapper(self, *args, **kwargs):
    #         try:
    #             self.__connect()
    #             return True, func(self, *args, **kwargs)
    #         except Exception as err:
    #             logger.exception(str(err))
    #             return False, str(err)
    #     return wrapper
    #


Template = BaseTemplate



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：各种自定义控件
# 建立日期：2023.07.20
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-07-20  第一次编写
#

from os import remove
from csv import DictReader
from decimal import Decimal
from tempfile import NamedTemporaryFile
from os.path import exists

from .baseuiawrapper import BaseUIAWrapper
from ..error import RecordNotFoundError, RecordAmbiguousError, ItemKeyError, TimeoutError


class PromptWrapper(BaseUIAWrapper):

    _control_types = ['Prompt']

    def __init__(self, elem):
        super(PromptWrapper, self).__init__(elem)

    def __wait_prompt_close(self):
        try:
            # NOTE 使用_get_control从顶层窗口查找
            self._get_control({'handle': self.handle}).wait_not('exists')
        except TimeoutError:
            # 超时因存在关闭确认框或其他已知的原因
            pass

    @property
    def title(self):
        title_spec = self.child(self._client.PROMPT_TITLE_ID)
        return title_spec.window_text() if title_spec.exists() else ''

    def content(self):
        text_spec = self.child(self._client.PROMPT_CONTENT_ID)
        return text_spec.window_text() if text_spec.exists() else ''

    def ok(self):
        ok_btn = self.child({
            'title_re': self._client.PROMPT_OKBUTTON_TITLE,
            'control_type': 'Button'
        })
        if ok_btn.exists():
            ok_btn.click()
            self.__wait_prompt_close()

    def cancel(self):
        cancel_btn = self.child({
            'title_re': self._client.PROMPT_CANCELBUTTON_TITLE,
            'control_type': 'Button'
        })
        if cancel_btn.exists():
            cancel_btn.click()
            self.__wait_prompt_close()

    def close(self):
        # FIXME 有弹框关闭时会弹出确认对话框
        criterias = list(self._client.PROMPT_CLOSE_BUTTON)
        criterias.extend([
            {'title_re': self._client.PROMPT_CANCELBUTTON_TITLE, 'control_type': 'Button'},
            {'title_re': self._client.PROMPT_OKBUTTON_TITLE, 'control_type': 'Button'}
        ])

        for criteria in criterias:
            cls_btn = self.child(criteria)
            if cls_btn.exists():  # 非捕捉模式
                cls_btn.click()
                self.__wait_prompt_close()
                break


class GridItem(object):
    """表格中的项，非控件"""

    def __init__(self, grid, data):
        self.__grid = grid
        self.__data = data

        config = self.__grid.config
        self.__headHeight = 24 if config('headHeight') is None else config('headHeight')
        self.__lineHeight = 24 if config('lineHeight') is None else config('lineHeight')
        self.__offset = 6 if config('offset') is None else config('offset')

    def __getitem__(self, item):
        try:
            return self.__data[item]
        except KeyError:
            raise ItemKeyError(f'表格中没有<{item}>字段')

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)

    def __str__(self):
        return str(self.__data)

    def __repr__(self):
        return str(self.__data)

    def __getattribute__(self, attr):
        return object.__getattribute__(self, attr)

    def __getattr__(self, item):
        if item not in ['pop', 'popitem', 'update', 'setdefault', 'clear', 'fromkeys']:
            return getattr(self.__data, item)
        else:
            raise AttributeError(f'GridItem对象没有{item}属性')

    def click(self, x=None, double=False):
        self.__grid.click_input(
            coords=(x, self.__headHeight + int(self.__lineHeight >> 1) + (self.__lineHeight * self.__data['index'])),
            double=double
        )

    def double_click(self):
        self.click(double=True)

    def select(self):
        self.click(x=self.__offset)


class GridWrapper(BaseUIAWrapper):

    _control_types = ['GridCSV']

    def __init__(self, elem):
        super(GridWrapper, self).__init__(elem)

    def __getitem__(self, item):
        return self.__data[item]

    def __getattribute__(self, attr):
        return object.__getattribute__(self, attr)

    def __getattr__(self, item):
        if item in ['count', 'index', 'copy']:
            return getattr(self.__data, item)
        else:
            raise AttributeError(f'GridWrapper对象没有{item}属性')

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)

    def __repr__(self):
        rtn = '['
        for item in self.__data:
            rtn += '\n\t' + str(item) + ','
        return rtn + '\n]'

    def __str__(self):
        rtn = '['
        for item in self.__data:
            rtn += '\n\t' + str(item) + ','
        return rtn + '\n]'

    def __saveto(self, file):
        # 关闭可能存在的弹窗
        self._prompt.close()
        self.set_focus().type_keys('^s')

        saveto = self._get_control(self.config('saveto'))
        # saveto.child(self.config('savetofile')).set_text(file)
        savetofile = saveto.child(self.config('savetofile'))
        # 将鼠标移动到输入框，否则微软UIA的接口会找不到主窗口，不知何故
        savetofile.click_input()
        savetofile.set_text(file)
        saveto.ok()

    def __save_csv_and_parse(self):
        """使用另存为方式保存数据"""
        with NamedTemporaryFile(mode='w+', prefix='WYH_', suffix='.csv', newline='', delete=True) as f:
            file = f.name

        self.__saveto(file)
        while not exists(file):  # 等待保存完成
            pass

        with open(file, newline='') as csvfile:
            reader = DictReader(csvfile)
            self.__data = [GridItem(self, dict(index=reader.line_num-2, **row)) for row in reader]  # row为何是str？

        if exists(file):
            remove(file)

    def items(self, **kwargs):
        """
        依据给定的条件过滤列表，返回过滤后的列表（行，即GridItem对象）

        kwargs关键字可以是表格标头的任何一个字段，value是一个字符串或由字符串组成的元组，
        即使像成交价格、成交数量等在GridWrapper中仍然以字符串格式保存，这样做的好处是
        便于使用Decimal类进行浮点数运算，而不会因计算机浮点数危机使价格计算错误。

        items()方法是GridWrapper对象的核心方法，使用场景可能如下：

        1、获得全部委托单
        grid.items()
        2、使用一个关键字参数过滤列表
        grid.items(证券名称='农业银行')  # 所有证券名称为‘农业银行’的委托单
        3、使用多个关键字参数过滤列表
        grid.items(证券名称='农业银行', 操作='买入')  # 将农业银行的买入单过滤出来
        4、使用一个关键字参数，多值过滤列表
        grid.items(证券名称=('农业银行', '平安银行'))  # 所有证券名称为‘农业银行’和‘平安银行’的委托单
        grid.items(合同编号=('123456', '654321'))  # 合同编号为‘123456’和‘654321’的委托单
        5、使用多关键字参数，多值过滤列表
        grid.items(证券名称=('农业银行', '平安银行'), 操作='买入')  # 农业银行和平安银行的买入单
        """
        table = self.__data.copy()
        for key, value in kwargs.items():
            values = (str(value),) if isinstance(value, (str, int, float, Decimal)) else value
            table = [row for row in table if row[key] in values]
        return table

    def item(self, **kwargs):
        """依据给定的条件，返回一个匹配的项目"""
        table = self.items(**kwargs)

        if not table:
            raise RecordNotFoundError(kwargs)

        if len(table) > 1:
            exception = RecordAmbiguousError('有{0}条记录, 在此条件下{1}'.format(len(table), str(kwargs),))
            exception.table = table
            raise exception

        return table[0]

    def refresh(self):
        self.type_keys('{F5}')
        # FIXME 等待刷新完成？self是否还有效？
        self.__save_csv_and_parse()
        return self


class JQKGridWrapper(BaseUIAWrapper):
    """为同花顺客户端定制的表格控件，因其保存时会弹出验证码"""

    _control_types = ['JQKGridCSV']

    def __init__(self, elem):
        super(JQKGridWrapper, self).__init__(elem)

    def __getitem__(self, item):
        return self.__data[item]

    def __getattribute__(self, attr):
        return object.__getattribute__(self, attr)

    def __getattr__(self, item):
        if item in ['count', 'index', 'copy']:
            return getattr(self.__data, item)
        else:
            raise AttributeError(f'THSGridWrapper对象没有{item}属性')

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)

    def __repr__(self):
        rtn = '['
        for item in self.__data:
            rtn += '\n\t' + str(item) + ','
        return rtn + '\n]'

    def __str__(self):
        rtn = '['
        for item in self.__data:
            rtn += '\n\t' + str(item) + ','
        return rtn + '\n]'

    def __saveto(self, file):
        # 关闭可能存在的弹窗
        self._prompt.close()
        self.set_focus().type_keys('^s')

        saveto = self._get_control(self.config('saveto'))

        # 捕捉是否有验证码提示框
        pane = self._prompt.tooltip(text=self.config('verifycode_text'))
        if pane is not None:
            # 由于提示框tooltip的捕捉机制以及弹窗的等待关闭机制，故exists()不需要捕捉模式
            while not saveto.exists():  # 默认非捕捉模式
                pane.child(self.config('verifycode_id')).set_text(
                    # 选择是否主动刷新验证码
                    # pane.child(self.config('verifycodeimage_id').refresh().window_text()
                    pane.child(self.config('verifycodeimage_id')).window_text()
                )
                # 由于点击确认后，验证码提示框并不会主动关闭，故判断“另存为”对话框是否弹出作为验证码是否正确的判断依据
                pane.ok()
            else:
                # 关闭验证码提示框
                pane.cancel()

        savetofile = saveto.child(self.config('savetofile'))
        # 将鼠标移动到输入框，否则微软UIA的接口会找不到主窗口，不知何故
        savetofile.click_input()
        # 这里感觉“卡”一下，是因为虽然“另存为”对话框已弹出，但弹窗的“等待关闭”机制仍在等待验证码提示框关闭，但验证码提示框不关闭
        savetofile.set_text(file)
        # 保存
        saveto.ok()

    def __save_csv_and_parse(self):
        """使用另存为方式保存数据"""
        with NamedTemporaryFile(mode='w+', prefix='WYH_', suffix='.csv', newline='', delete=True) as f:
            file = f.name

        self.__saveto(file)
        while not exists(file):  # 等待保存完成
            pass

        with open(file, newline='') as csvfile:
            reader = DictReader(csvfile, dialect='excel-tab')  # 同花顺表格采用excel-tab变种
            self.__data = [GridItem(self, dict(index=reader.line_num-2, **row)) for row in reader]  # row为何是str？

        if exists(file):
            remove(file)

    def items(self, **kwargs):
        """
        依据给定的条件过滤列表，返回过滤后的列表（行，即GridItem对象）

        kwargs关键字可以是表格标头的任何一个字段，value是一个字符串或由字符串组成的元组，
        即使像成交价格、成交数量等在GridWrapper中仍然以字符串格式保存，这样做的好处是
        便于使用Decimal类进行浮点数运算，而不会因计算机浮点数危机使价格计算错误。

        items()方法是GridWrapper对象的核心方法，使用场景可能如下：

        1、获得全部委托单
        grid.items()
        2、使用一个关键字参数过滤列表
        grid.items(证券名称='农业银行')  # 所有证券名称为‘农业银行’的委托单
        3、使用多个关键字参数过滤列表
        grid.items(证券名称='农业银行', 操作='买入')  # 将农业银行的买入单过滤出来
        4、使用一个关键字参数，多值过滤列表
        grid.items(证券名称=('农业银行', '平安银行'))  # 所有证券名称为‘农业银行’和‘平安银行’的委托单
        grid.items(合同编号=('123456', '654321'))  # 合同编号为‘123456’和‘654321’的委托单
        5、使用多关键字参数，多值过滤列表
        grid.items(证券名称=('农业银行', '平安银行'), 操作='买入')  # 农业银行和平安银行的买入单
        """
        table = self.__data.copy()
        for key, value in kwargs.items():
            values = (str(value),) if isinstance(value, (str, int, float, Decimal)) else value
            table = [row for row in table if row[key] in values]
        return table

    def item(self, **kwargs):
        """依据给定的条件，返回一个匹配的项目"""
        table = self.items(**kwargs)

        if not table:
            raise RecordNotFoundError(kwargs)

        if len(table) > 1:
            exception = RecordAmbiguousError('有{0}条记录, 在此条件下{1}'.format(len(table), str(kwargs),))
            exception.table = table
            raise exception

        return table[0]

    def refresh(self):
        self.type_keys('{F5}')
        # FIXME 等待刷新完成？self是否还有效？
        self.__save_csv_and_parse()
        return self



#
# 券商客户端自动化测试库
# Copyright (C) 2023 谁的谁（41715399@qq.com） All rights reserved.
#
# 模块功能：管理各类弹窗和提示窗
# 建立日期：2022.12.01
# 联系方式：谁的谁（41715399@qq.com）
#
# 开源软件声明：
# 本软件遵守“MIT License”开源协议开源，仅供学习和参考。您可以自由使用或修改源代码或二进制文件，但必须保留上述版权声明。
# 该软件旨在深度学习和挖掘python pywinauto库的功能和潜力，由于环境的不确定性和该软件的不可靠性，请不要将该软件应用于
# 实盘交易。如您确需量化交易实盘功能，请使用券商提供的量化交易平台，否则由于您使用该软件实盘交易所造成的账户损失或政策风
# 险，开源软件提供者或插件提供者均不承担任何责任。同时，无论是直接的、间接的、偶然的、潜在的因使用该软件所造成的账号安全
# 损失、数据安全损失、账户资产损失或其他任何责任事故，开源软件提供者或插件提供者均不承担任何责任。请不要将该软件应用于商
# 业活动，否则由于把该软件应用于商业活动所造成的一切损失或法律责任，开源软件提供者或插件提供者均不承担任何责任。
#
# 修改日志：
#   2022-12-01  第一次编写
#

import re
import operator
import time

from threading import Thread
from pywinauto.timings import wait_until_passes, TimeoutError, Timings
from pywinauto.findbestmatch import find_best_control_matches

from .control.wrappersa import PromptWrapper


class PromptManagerMeta(type):

    def __init__(cls, name, bases, attrs):
        super(PromptManagerMeta, cls).__init__(name, bases, attrs)

    def __call__(cls, client):
        if client.prompt is None:
            client.prompt = super(PromptManagerMeta, cls).__call__(client)
        return client.prompt


class PromptManager(metaclass=PromptManagerMeta):
    """弹出框、提示框管理器"""

    def __init__(self, client):
        self.__client = client
        self.__monitorthread = None
        self.__monitorrun = False

    def __get_prompts(self):
        """获得所有弹窗"""
        window = self.__client.window()
        child_count = window.child_count  # 似乎以计数的方法更为可靠

        elements = window.element_info.children()[:-child_count]  # 需要返回元素，而非包装器

        if elements:
            elements = [elem for elem in elements if elem.control_type in self.__client.PROMPT_TYPE_LIST]

        if elements:
            elements = [elem for elem in elements if len(elem.children(process=elem.process_id)) > 0]

        # 返回弹窗包装器
        return [PromptWrapper(ele) for ele in elements]

    def __find_prompts(self,
                       title=None,       # 对话框标题，支持正则表达式
                       content=None,     # 对话框内容
                       text=None,        # **对话框中的所有inspect.exe可见文字字符串，支持正则表达式，这是一个万能参数**
                       best_match=None,  # pywinauto的参数，可以用但作用不大
                       func=None,        # 定义一个函数去筛选
                       ):
        """依据给定的条件，筛选符合条件的对话框"""

        panes = self.__get_prompts()

        if title is not None and panes:
            panes = [pane for pane in panes if re.match(title, pane.title)]

        if content is not None and panes:
            panes = [pane for pane in panes if re.match(content, pane.content())]

        if text is not None and panes:
            panes = [pane for pane in panes if list(filter(lambda x: re.match(text, x), pane.texts()))]

        if best_match is not None and panes:
            panes = find_best_control_matches(best_match, panes)

        if func is not None and panes:
            panes = [pane for pane in panes if func(pane)]

        return panes

    def __monitor(self, duration, kwargs):
        """监视弹窗并关闭"""
        start = time.perf_counter()

        while self.__monitorrun is True:
            for pane in self.__find_prompts(**kwargs):
                pane.close()

            time_left = duration - (time.perf_counter() - start)

            if time_left > 0:
                time.sleep(min(Timings.window_find_retry, time_left))
            else:
                self.__monitorrun = False

    def stop_monitor(self):
        if self.__monitorthread is not None and self.__monitorthread.is_alive():
            self.__monitorrun = False
            self.__monitorthread.join()

    def start_monitor(self, delay=0, **kwargs):
        self.stop_monitor()
        self.__monitorrun = True
        self.__monitorthread = Thread(target=self.__monitor, name='PromptClose', args=(delay, kwargs))
        self.__monitorthread.start()

    def close(self, **kwargs):
        """关闭当前所有存在的弹窗"""
        self.stop_monitor()
        for pane in self.__find_prompts(**kwargs):
            pane.close()

    def tooltip(self, timeout=None, retry_interval=None, **kwargs):
        """提示框应用场景，采用捕捉模式"""
        self.stop_monitor()
        # 不要设置为默认值
        if timeout is None:
            timeout = Timings.window_find_timeout * 0.6
        if retry_interval is None:
            retry_interval = Timings.window_find_retry

        try:
            return wait_until_passes(timeout,
                                     retry_interval,
                                     operator.getitem,
                                     (IndexError,),
                                     self.__find_prompts(**kwargs),
                                     0  # 总是返回最上面的提示框
                                     )
        except TimeoutError:
            return None

    def exists(self, timeout=None, retry_interval=None, **kwargs):
        """判断弹窗是否存在，默认非捕捉模式"""
        # 不要设置为默认值，时间会更改
        if timeout is None:
            # timeout = Timings.exists_timeout
            timeout = 0
        if retry_interval is None:
            retry_interval = Timings.exists_retry

        return self.tooltip(timeout, retry_interval, **kwargs) is not None
